#!/bin/env python
# -*- coding: utf-8
# vim: set fileencoding=utf-8

#    Copyright (C) 2008-2020 Thomas Capricelli <orzel@freehackers.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# System
import sys, logging, logging.handlers, email, traceback
from email.parser import Parser
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTPException
from signal import SIGTERM, SIGINT, signal
from optparse import make_option
from time import sleep
from datetime import date, datetime
from random import choice
from os import getpid, path, environ, system, unlink, listdir, stat, mkdir, umask, devnull
from shutil import move
from subprocess import call, STDOUT
FNULL = open(devnull, 'w')

# Django
from django.core.management.base import BaseCommand, CommandError
from django.core.mail import get_connection, mail_admins
from django.conf import settings
from django.db import connection

# Project
from main.models import *
from main.views import find_email
from main.forward import create_outbound_message
from main.helpers import create_random_string

ARCHIVE_DIR=path.abspath(settings.ARCHIVE_DIR)
BOUNCED_DIR=path.abspath(settings.BOUNCED_DIR)

# interval at which the mainloop wakes up to check everything (in seconds)
WAKEUP_INTERVAL = 5

#
# Logging
#

# https://docs.djangoproject.com/en/1.7/ref/django-admin/#django-admin-option---verbosity
def deepdebug(message):
    if verbosity>2: log("DeepDBG", message)
def debug(message):
    if verbosity>1: log("DEBUG", message)
def error(message):
    if verbosity>0: log("\x1B[31mERROR\x1B[0m",  message)
def error_passon(message):
    error(message)
    mail_admins(
        "Important error from the colibri daemon",
        "The following error was reported and requires your attention:\n\n\t%s" % message,
        fail_silently=True, # no need to pile up exceptions
    )
def info(message):
    if verbosity>0: log("INFO", message)
def log(level, message):
    stdout.write("%s %7s %s" % (
        datetime.now().strftime("%Y-%m-%d %H:%M"),
        level,
        message,
    ))

    stdout.flush()

#
# utilities
#

def move_file(oldpath, _newpath, msg="moving"):
    # check if destination already exists
    newpath = _newpath
    while path.exists(newpath):
        newpath = _newpath+'.'+create_random_string(7)
    deepdebug('%s %s -> %s' % (msg, oldpath, newpath))
    move(oldpath, newpath)
    return

def checkAndCreateDirectory(dirname,basedir):
    """
    Ensure the whole path exists, creating missing directories
    This is not really bound to the instance/class, and could/should
    be moved outside of the class.
    """
    # cleaning
    dirname = path.normpath(dirname)
    basedir = path.normpath(basedir)
#        deepdebug('checkAndCreateDirectory(%s)'% dirname)
    # check for both bad argument or problem with recursion
    assert(dirname.startswith(basedir))
    try:
        stat(dirname)
    except OSError:
        # does not exist, create it
        deepdebug('"%s" does not exist, create it' % dirname)
        checkAndCreateDirectory(path.dirname(dirname),basedir) # recursive check of parent dir
        info("Creating %s" % dirname)
        mkdir(dirname)
    return


def is_subscriber(fromfield, mailinglist):
    """
    Check if the fromfield is a subscriber of the list, taking
    OtherEmailKey into account.

    Return True of False.
    """
    # find out email
    name, address = email.utils.parseaddr(fromfield) # split "myname<myemail>" -> [myname, myemail]
    if address=='':
        info("is_subscriber : Can't extract email in the 'From' field ('%s')" % fromfield)
        return False

    # find profile
    deepdebug("is_subscriber with address = %s" % address)
    profile = find_email(address)
    if profile is None: return False

    # find
    count = Subscription.objects.filter(profile=profile, mailinglist=mailinglist).count()
    deepdebug("is_subscriber : profile found, count is %d" % count)
    return count>0

class Colibrid:
    """
    implementation of the daemon
    """

    def __init__(self):
        # install signal handler
        signal(SIGTERM, self.catch_sig)
        signal(SIGINT, self.catch_sig)
        self.stop_asap = False
        # TODO : probably we should use HeaderParser here
        self.parser = Parser()

    def catch_sig(self, sig, stack):
        info('Caught signal %d' % sig)
        self.stop_asap = True

    def initial_checks(self):
        """
        Perform some initial checks, such as
        * can we access the incoming dir ?
        * can we unlink files in the incoming dir ?
        * can we write to the reject dir ?
        * can we write to the archve dir ?
        * can we find mhonarc ?
        * can we execute mhonarc ?

        * r/w access to pid file
        * r/w access to rw/ dir
        * r/w access do dataase

        This method is responsible for outputting useful error() messages
        """
        # check that mhonarc is available
        if call(["which", "mhonarc"], stdout=FNULL, stderr=STDOUT)!=0:
            error("mhonarc can not be found")
            return False
        debug("mhonarc found")
        if call(["mhonarc", "--help"], stdout=FNULL, stderr=STDOUT)!=0:
            error("mhonarc can not be run")
            return False
        # check that important directories are available
        if not path.exists(settings.INCOMING_DIR):
            error('INCOMING_DIR does not exist : "%s"' % settings.INCOMING_DIR)
            return False
        if not path.exists(settings.REJECT_DIR):
            error('REJECT_DIR does not exist : "%s"' % settings.REJECT_DIR)
            return False
        if not path.exists(BOUNCED_DIR):
            error('BOUNCED_DIR does not exist : "%s"' % BOUNCED_DIR)
            return False
        if not path.exists(ARCHIVE_DIR):
            error('ARCHIVE_DIR does not exist : "%s"' % ARCHIVE_DIR)
            return False
        # TODO..
        return True

    def run(self):
        deepdebug("colibrid run() called")
        while not self.stop_asap:
            self.checkIncomingMails()
            # Close the connection so as not to be caught by a timeout
            # as recommended from https://code.djangoproject.com/ticket/21597#comment:29
            # close it _before_ waiting and spending time
            connection.close()
            sleep( WAKEUP_INTERVAL )

    def checkIncomingMails(self):
        files = listdir(settings.INCOMING_DIR)
        # ignore some common patterns
        for n in [ '.keep', 'old', 'save' ]:
            while n in files: files.remove(n)
        if len(files)<1: return
        deepdebug('checkIncomingMails : %d file(s)'%len(files))
        for filename in files:
            # handle it
            self.checkOneFile(filename)
            # check the file is removed, whatever the way
            # (forwarded/archived, rejected, error, bounce...)
            assert not path.exists(path.join(settings.INCOMING_DIR,filename))

    def checkOneFile(self, filename):
        # basic checks
        deepdebug('checkOneFile : %s' % filename)
        with open(path.join(settings.INCOMING_DIR, filename), 'r') as f:
            try:
                m = self.parser.parse(f)
            except:
                info("Error while parsing file %s: %s, mail rejected" % (filename, str(traceback.format_exc())))
                self.reject(filename)
                return
        if not self.checkMailValidity(m):
            info("The content of %s is not valid, mail rejected" % filename)
            self.reject(filename)
            return
        # Find out recipient / clean
        if m['X-Original-To'] is not None:
            recipient = m['X-Original-To']
        else:
            recipient = m['Delivered-To']
        assert(recipient is not None)
        name, addr = email.utils.parseaddr(recipient) # split "myname<myemail>" -> [myname, myemail]
        if addr=='':
            info("Can't extract email from recipient '%s', rejected" % recipient)
            self.reject(filename)
            return
        name, host = addr.split('@')
        if name=='' or host=='':
            info("Can't extract name and host from recipient '%s', got '%s','%s', rejected" % (recipient, name, host))
            self.reject(filename)
            return
        deepdebug('To: %s,%s' % (name, host))

        # Check for *-editor *-subscribe *-unsubscribe
        if name.endswith('-editor') or name.endswith('-subscribe') or name.endswith('-unsubscribe') or name.endswith('-help'):
            # TODO : improve speed with regexp here ?
            info("-editor/-subscribe/-unsubsribe/-help not handled yet, mail rejected")
            self.reject(filename)
            return

        # check for *-request
        if name.endswith('-request'): # TODO : improve speed with regexp here ?
            self.handleRequest(name[:-8], host, m, filename)
            return

        # check for *-owner
        if name.endswith('-owner'): # TODO : improve speed with regexp here?
            self.handleOwner(name[:-6], host, m, filename)
            return

        # ok, now, the mail should be toward an actual mailing list

        # Try to find the mailing list
        lists = List.objects.filter(name=name).filter(host=host)
        deepdebug('Found %d lists for this criteria' % lists.count())
        if lists.count()!=1:
            info("Unknown mailing list : %s@%s, rejected" % (name,host))
            self.reject(filename)
            return

        # ok, now we have the corresponding mailing list
        mailinglist = lists[0] # dont call the variable 'list', it's the name of a python class
        info('Incoming mail for %s' % mailinglist)

        # check list state
        if mailinglist.state != LIST_STATE_PRODUCTION:
            # TODO : send an email back for all those errors, but not until
            # loop detection / bounces and such are implemented
            info("This mailing list is not in production state, rejecting")
            self.reject(filename)
            return

        # check this mail is allowed to get through
        if is_subscriber(m['From'], mailinglist):
            # TODO : handle MODERATE case
            accepted = mailinglist.posting_from_subscriber == LIST_MODERATE_ACCEPT
            error = "Subscribers are not allowed to post to the list (read-only list)"
        else:
            # TODO : handle MODERATE case
            accepted = mailinglist.posting_from_external == LIST_MODERATE_ACCEPT
            error = "People not subscribed to the list are not allowed to post."
        if not accepted:
            # TODO: probably better to keep a copy in rw/refused/<list>
            unlink(path.join(settings.INCOMING_DIR,filename))
            self.refuse(m, mailinglist, error)
            return
        self.forward_and_archive(m, mailinglist, filename)

    def forward_and_archive(self, m, mailinglist, filename):
        deepdebug('forward_and_archive for list %s' % str(mailinglist))
        #
        # find active subscribers, exit if none
        #
        subscribers = [sub.profile.user.email for sub in Subscription.objects.filter(mailinglist=mailinglist).filter(profile__user__is_active=True)]
        nbsubscribers = len(subscribers)
        if nbsubscribers==0:
            info("No active subscriber for %s, done." % str(mailinglist))
            return
        debug("Found %d active subscribers" % nbsubscribers)
        deepdebug("active subscribers : '%s'" % ", ".join(subscribers))

        # send
        mail = create_outbound_message(m, mailinglist)
        self.forward_mail_to_subscribers(mail, subscribers, mailinglist)

        # archive
        archive_dir = self.archive_mail(filename, mailinglist)
        # TODO: currently we start mhonarc once for every mail, which is clearly
        # not optimized
        self.do_mhonarc(archive_dir, str(mailinglist))

    def forward_mail_to_subscribers(self, mail, subscribers, mailinglist):
        # actually forward it
        try:
            connection = get_connection() # we use the django one, configured with stuff from settings
            connection.open()

            if not connection.connection:
                raise SMTPException("Fail to create SMTPConnection.connection")
            connection.connection.sendmail(
                mailinglist.list_address_extended('owner'),  # from, will be set in Return-Path
                subscribers,                            # to
                mail.as_string())                       # content
        except SMTPException as detail:
            error_passon('SMTP Error while sending the email: %s' % str(detail))
            return # TODO : prevent loops if smtp fails : the mail stays in incoming and will be tried again in loop
        info("Mail forwarded to %d subscribers" % len(subscribers))

    def refuse(self, m, mailinglist, reason):
        deepdebug('Refuse mail for %s' % str(mailinglist))

        msgfrom = mailinglist.list_address_extended('owner')
        msgto = m['From']
        reason = "\n\nYour message to the list %s has been refused.\n%s\n" % (str(mailinglist), reason)

        # create mail
        mail = MIMEMultipart()
        mail['Subject'] = "Your message to the list %s has been refused " % str(mailinglist)
        mail['From'] = msgfrom
        mail['To'] = msgto
        mail.attach(MIMEText(reason, 'plain', 'utf8')) # error msg
        mail.attach(m) # previous mail

        # ugly hack we need to add headers at the TOP
#        items = m.items()
#        for key,val in items: m.__delitem__(key) # delete them all
#        for key,val in items: m.add_header(key,val) # re-add them all
#        mail.attach(m) # add original mail

        # send it
        try:
            connection = get_connection() # we use the django one, configured with stuff from settings
            connection.open()
            if not connection.connection:
                raise SMTPException("Fail to create SMTPConnection.connection")
            connection.connection.sendmail(
                mailinglist.list_address_extended('owner'),  # from, will be set in Return-Path
                [ msgto ],                              # to
                mail.as_string())                       # content
        except SMTPException as detail:
            error_passon('refuse: SMTP Error while sending the email: %s' % str(detail))
            return # TODO : prevent loops if smtp fails : the mail stays in incoming and will be tried again in loop
        debug("Refused mail sent to %s" % msgto)

    def archive_mail(self, filename, mailinglist):
        """
        store a (raw) copy of the email in the right archive
        directory.
        returns the corresponding (main) archive directory

        the main archive directory is ARCHIVE_DIR/<listname>/<year>/<month>/
        there are two directories inside this one
            /text/ for the original raw mails
            /html/ for the html stuff generated by mhonarc
        """

        today = date.today()
        date.today().year
        archive_dir = path.join(ARCHIVE_DIR, str(mailinglist), "%s/%s" % (today.year, today.month))
        text_dir = path.join(archive_dir, "text")
        html_dir = path.join(archive_dir, "html")
#        deepdebug("archive_mail, archive_dir is %s" % archive_dir)
        # check the directory exists
        checkAndCreateDirectory(text_dir, ARCHIVE_DIR)
        # move the file
        move_file(path.join(settings.INCOMING_DIR,filename), path.join(text_dir,path.basename(filename)), 'archive')
        return archive_dir

    def do_mhonarc(self, archive_dir, list_name):
        """
        Call mhonarc to update the archive : html files are created from new raw files.
        """
        text_dir = path.join(archive_dir, "text")
        html_dir = path.join(archive_dir, "html")
        checkAndCreateDirectory(html_dir, ARCHIVE_DIR)
        title = "Archive for the list %s" % list_name
        ttitle = "Threaded archive for the list %s" % list_name
        resourcefile = path.join(settings.BASE_DIR, "scripts", "myresource.mrc")
        command = 'mhonarc -mhpattern "^[^\\.]" -add -quiet -title "%s" -ttitle "%s" -rcfile %s -outdir %s %s' % (title, ttitle, resourcefile, html_dir, text_dir)
        debug('do_mhonarc() : executing "%s"' % command)
        errno = system(command)
        if errno!=0:
            error_passon("do_mhonarc(%s,%s) : error %d while executing mhonarc" % (html_dir, list_name, errno))

    def handleRequest(self, name, host, m, filename):
        """
        name is the name without the -request suffix
        m is an instance of email.message.Message
        """
        info("-request not handled yet, mail rejected")
        self.reject(filename)
        return

    def handleOwner(self, name, host, m, filename):
        """
        name is the name without the -owner suffix
        m is an instance of email.message.Message
        """

        # Try to find the mailing list
        lists = List.objects.filter(name=name).filter(host=host)
        if lists.count()!=1:
            info("handleOwner: unknown mailing list : %s@%s, rejected" % (name,host))
            self.reject(filename)
            return
        mailinglist = lists[0]

        # find out the report part
        if not m.get_content_type()=='multipart/report':
            info("handleOwner : returned message is not a multipart/report, don't know what to to with it, rejected")
            self.reject(filename)
            return
        for part in m.walk():
            if part.get_content_type()=='message/delivery-status':
                break
        else:
            info("handleOwner : Can't find the message/delivery-status part in returned message, rejected")
            self.reject(filename)
            return

        # Following http://www.faqs.org/rfcs/rfc3464.html (error msg format)
        # and  http://www.ietf.org/rfc/rfc3463.txt (status code)

        # find out the problematic address
        for line in part.as_string().split('\n'):
            if line.startswith('Final-Recipient:'):
                break
        else:
            info("handleOwner : Can't find the Final-Recipient: in deliveery-status, rejected")
            self.reject(filename)
            return
        bounced_email = line.split(':')[1].split(';')[1].strip()
        info("handleOwner : bounced email is %s" % bounced_email)

        # find out the status string
        for line in part.as_string().split('\n'):
            if line.startswith('Status:'):
                status = line.split(':')[1].strip()
                break
        else:
            status = "Colibri could not find the status sting the the bounced mail"

        # find the colibri profile associated with this email
        profile = find_email(bounced_email, checkOthers=False)
        if profile == None:
            info("handleOwner : Can't find this email in the database, returned mail rejected")
            self.reject(filename)
            return

        # create bounce object, update profile
        bounce = Bounce(profile=profile, status=status, filename= "/%s/%s" % (str(mailinglist),filename)) # TODO :  parse and set bounce.datetime
        bounce.save()
        profile.bounces_count = profile.bounces_count+1
        profile.save()

        # move the file to rw/bounced/*
        # specification : the file goes to the directory BOUNCED_DIR/<listname>/
        bounced_dir = path.join(BOUNCED_DIR, str(mailinglist))
        checkAndCreateDirectory(bounced_dir,BOUNCED_DIR)
        move_file(path.join(settings.INCOMING_DIR,filename), path.join(bounced_dir, path.basename(filename)), "bounced file")
        # we are done

    def checkMailValidity(self, mail):
        """
        Apply some heuristics to decide if the mail is valid (return True),
         or not (return False)
        """
        if mail['From']==None: return False
        if mail['Date']==None: return False
        if mail['Message-ID']==None: return False
        if mail['X-Original-To']==None and mail['Delivered-To']==None: return False
        return True

    def reject(self, filename):
        """
        Reject the mail to some folder
        """
        move_file(path.join(settings.INCOMING_DIR,filename), path.join(settings.REJECT_DIR,path.basename(filename)), 'reject')

class Command(BaseCommand):
    help = """
    colibrid is the daemon in charge of checking incoming mails in the right
    directory, and handle them, which means:
        * check for bounces
        * create archives/html files
        * forward mails to subscribed people
    """
    leave_locale_alone = True

    def add_arguments(self, parser):
        parser.add_argument("--debug", action="store_true", dest="debug", default=False, help="only does one check and exit"),

    def handle(self, *args, **options):
        # Set global variables
        global verbosity, stdout
        verbosity = int(options['verbosity'])
        stdout = self.stdout

        if options['debug']:
            # do only one check, could be used with a cron
            colibrid = Colibrid()
            if colibrid.initial_checks():
                colibrid.checkIncomingMails()
            sys.exit(0)

        # start the daemon main loop
        info('Colibrid started, watching INCOMING_DIR=%s'% settings.INCOMING_DIR)
        colibrid = Colibrid()
        if colibrid.initial_checks():
        #    colibrid.run()
            try:
                colibrid.run()
            except: # python doc says it is a very bad practice to catch all exception
                error_passon("Unexpected error : %s" % str(traceback.format_exc()))
                info('Caught something while running mainloop() : giving up')
        info('colibrid exiting cleanly')

# vim: ai ts=4 sts=4 et sw=4

