#
# -*- coding: utf-8
# vim: set fileencoding=utf-8

#    Copyright (C) 2008-2011 Thomas Capricelli <orzel@freehackers.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# System
from os.path import abspath
from os import stat, listdir
from platform import python_version
from calendar import month_name

# django
from django.urls import reverse
from django.shortcuts import get_object_or_404, render
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_POST
from django.http import HttpResponseRedirect, Http404, HttpResponseServerError
from django.contrib.sites.models import Site
from django.contrib.auth import get_user_model, login
from django.template.loader import render_to_string
from django.template import loader
from django.core.mail import send_mail
from django.conf import settings
from django import get_version
from django.dispatch import receiver
from django.contrib import messages

# Project
from registration.signals import user_activated
from main.models import *
from main.forms import EmailForm, ListForm

ARCHIVE_DIR=abspath(settings.ARCHIVE_DIR)

#
# helpers
#
def find_email(email, checkOthers=True, checkConfirmed=True):
    """
    This helper try to find the email given in argument either in the User
    table or in the OtherEmail table. It returns a Profile object if found,
    and None else.
    """
    # first look in User
    users = get_user_model().objects.filter(email=email)
    if users.count()>1:
        # TODO : should not happen
        raise ValueError()
    if users.count()>0:
        return users[0].profile

    # then in OtherEmail
    if not checkOthers: return None

    others = OtherEmail.objects.filter(email=email)
    if checkConfirmed: others = others.filter(confirmed=True)

    if others .count()>1: raise ValueError() # TODO : should not happen
    if others.count()>0:
        return others[0].profile

    # nothing found
    return None

def create_profile(user):
    profile = Profile()
    profile.user = user
    profile.save()
    return profile

#
# views
#
def credits(request):
    return render(request, 'credits.html', {
        'python_version': python_version(),
        'django_version': get_version(),
    })

def server_error(request, template_name='500.html'):
    t = loader.get_template(template_name)
    return HttpResponseServerError(t.render({
        "STATIC_URL" : settings.STATIC_URL,
    }))

def main(request):
    mailinglists = List.objects.filter(public=True).filter(state=LIST_STATE_PRODUCTION)
    return render(request, 'mainpage.html', {
        'mailinglists': mailinglists,
        'topics': Topic.objects.filter(list__public=True).distinct(),
    })

def display_topic_info(request, name):
    topic = get_object_or_404(Topic, name=name)
    mailinglists = List.objects.filter(public=True).filter(topic=topic).filter(state=LIST_STATE_PRODUCTION)
    return render(request, 'topic_info.html', {
        'topic': topic,
        'mailinglists':mailinglists
    })

def display_list_info(request, name):
    mailinglist = getListFromName(name)
    subscription = None
    profile = None
    if request.user.is_authenticated:
        profile = request.user.profile
        subscription = profile.list_subscription(mailinglist)
    su = request.user.is_authenticated and request.user.is_staff
    if not mailinglist.public and subscription == None and not su:
        raise Http404("private mailing list")
    # TODO: move this to model
    can_browse_subscribers = mailinglist.subscribers_list_visible or (subscription and subscription.moderator)
    form = EmailForm()
    return render(request, 'list_info.html', {
        'profile': profile,
        'mailinglist': mailinglist,
        'can_browse_archive': mailinglist.can_browse_archive(subscription!=None) or request.user.is_staff,
        'can_browse_subscribers': can_browse_subscribers or request.user.is_staff,
        'subscription': subscription,
        'form': form,
    })


@require_POST
@login_required
def list_unsubscribe(request, name):
    mailinglist = getListFromName(name)
    profile = request.user.profile
    subscription = profile.list_subscription(mailinglist)
    if subscription==None:
        return render(request, "error.html", {
            "message": "You are not subscribed to this list, how did you get there ?",
            "returnpage": "/list/%s/"%str(mailinglist),
        })
    # owner can not unsubscribe
    if mailinglist.owner.user==request.user:
        raise Http404("The owner of a mailing list can not unsubscribe")
    # actually unsubscribe
    subscription.delete()
    # send the resignation email
    email = profile.user.email
    subject = "You just left the mailing list %s" % str(mailinglist)
    message = render_to_string('bye_email.txt',
                               { 'mailinglist': mailinglist,
                                 'email' : email,
                                 'site': Site.objects.get_current() })
    send_mail(subject, message,  mailinglist.list_address_extended('owner'), [email])
    return render(request, 'unsubscribed.html', { 'mailinglist': mailinglist, })

def list_subscribe_authenticated(request, mailinglist):
    profile = request.user.profile
    subscription = profile.list_subscription(mailinglist)
    if subscription!=None:
        return render(request, "error.html", {
            "message": "You are already subscribed to this list, how did you get there ?",
            "returnpage": "/list/%s/"%str(mailinglist),
        })
    # ok, actually subscribe
    subscription = Subscription.objects.create(profile=profile, mailinglist=mailinglist)
    # send welcome email
    email = profile.user.email
    subject = "Welcome to the mailing list %s" % str(mailinglist)
    message = render_to_string('welcome_email.txt',
                               { 'mailinglist': mailinglist,
                                 'email' : email,
                                 'site': Site.objects.get_current() })
    send_mail(subject, message,  mailinglist.list_address_extended('owner'), [email])
    # done
    return render(request, 'subscribed.html', {'mailinglist': mailinglist, })

def list_subscribe_anonymous(request, mailinglist):
    form = EmailForm(request.POST)

    # check form
    if not form.is_valid():
        can_browse_archive = mailinglist.can_browse_archive(False)
        return render(request, 'list_info.html', {
            'profile':None,
            'mailinglist': mailinglist,
            'can_browse_archive': can_browse_archive,
            'subscription': None,
            'form': form,
        })

    # check the email is not already known
    email = form.cleaned_data['email']
    profile = find_email(email)
    if profile !=None:
        return render(request, "error.html", {
            "errorname":"Warning!",
            "message": """There is already an account using this email.
                If this is you, please identify yourself to proceed, if not, please choose another email to subscribe.
                If you have forgotten your password, you can <a href="/accounts/password/reset/">ask to change it here</a>.
            """,
             "returnpage" : "/accounts/login?next=/list/%s/"%str(mailinglist)
        })

    # first we need to create a user/profile
    from registration.models import RegistrationProfile
    # use a random password for now on, to be sure nobody can connect
    # the activation page will set the password
    password = get_user_model().objects.make_random_password()
    new_user = RegistrationProfile.objects.create_inactive_user(email=email, password=password, site=Site.objects.get_current(), send_email=False)
    profile  = create_profile(new_user)

    # send an email about the new account
    registration_profile= RegistrationProfile.objects.get(user=new_user)
    subject = "Confirmation for subscribing to list %s" % str(mailinglist)
    message = render_to_string('subscribe_email.txt',
                               { 'activation_key': registration_profile.activation_key,
                                 'expiration_days': settings.ACCOUNT_ACTIVATION_DAYS,
                                 'mailinglist': mailinglist,
                                 'email' : email,
                                 'password' : password,
                                 'site': Site.objects.get_current() })
    send_mail(subject, message,  mailinglist.list_address_extended('owner'), [email])

    # actual subscribing
    subscription = Subscription.objects.create(profile=profile, mailinglist=mailinglist)

    return render(request, 'subscribed_and_account_created.html', {'mailinglist': mailinglist, })

@require_POST
def list_subscribe(request, name):
    mailinglist = getListFromName(name)
    if request.user.is_authenticated:
        return list_subscribe_authenticated(request, mailinglist)
    else:
        return list_subscribe_anonymous(request, mailinglist)

# TODO : use regexp in urls.py instead
def getListFromName(name):
    """
    Try to find the list whose name is given as argument, raises
    a 404 if not found.
    """
    split = name.split('@')
    if len(split)!=2:
        raise Http404("This list does not exist.")
    return get_object_or_404(List.objects.filter(state=LIST_STATE_PRODUCTION), name=split[0], host=split[1])

def list_archives(request, name):
    mailinglist = getListFromName(name)
    #checks
    subscribed = request.user.is_authenticated and request.user.profile.list_subscription(mailinglist)!=None
    if not ( mailinglist.can_browse_archive(subscribed) or request.user.is_staff ):
        raise Http404("not allowed to browse archives")

    # find archives
    archive_dir = ARCHIVE_DIR+"/%s/" % str(mailinglist)
    try:
        years = listdir(archive_dir)
        years.sort()
    except OSError:
        # does not exist..
        return render(request, 'archives.html', {'mailinglist': mailinglist,})

    archives = []
    for year in years:
        months = listdir(archive_dir+year)
        months.sort(key=lambda m: int(m))
        archives.append ((year, months))
    return render(request, 'archives.html', {
        'mailinglist': mailinglist,
        'archives': archives,
    })

def list_subscribers(request, name):
    mailinglist = getListFromName(name)
    #checks
    subscription = None
    if request.user.is_authenticated:
        subscription = request.user.profile.list_subscription(mailinglist)
    can_browse_subscribers = mailinglist.subscribers_list_visible or (subscription and subscription.moderator)
    if not (can_browse_subscribers or request.user.is_staff):
        raise Http404("Not allowed to browse subscribers list")
    subscribers = [sub.profile for sub in Subscription.objects.filter(mailinglist=mailinglist).filter(profile__user__is_active=True)]
    return render(request, 'subscribers.html', {
        'mailinglist': mailinglist,
        'subscribers': subscribers,
    })

def list_archives_year(request, name, year):
    mailinglist = getListFromName(name)
    #checks
    assert(len(year)==4)
    assert(year.isdigit())
    subscribed = request.user.is_authenticated and request.user.profile.list_subscription(mailinglist)!=None
    if not ( mailinglist.can_browse_archive(subscribed) or request.user.is_staff ):
        raise Http404("not allowed to browse archives")

    # find archives
    archive_dir = ARCHIVE_DIR+"/%s/%s/" % (str(mailinglist), year)
    try:
        months = listdir(archive_dir)
        months.sort(key=lambda m: int(m))
    except OSError:
        # does not exist..
        return render(request, 'archives_year.html', {
            'mailinglist': mailinglist,
            'text': "No archive yet for this mailing list...",
        })

    archives = {}
    for month in months:
        assert(month.isdigit())
        archives[month] = len(listdir(archive_dir+month+'/text/'))

    return render(request, 'archives_year.html', {
        'mailinglist': mailinglist,
        'year': year,
        'archives': archives,
    })

month_numbers = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12']

def list_archives_month(request, name, year, month, path):
    mailinglist = getListFromName(name)
    #checks
    subscribed = request.user.is_authenticated and request.user.profile.list_subscription(mailinglist)!=None
    if not ( mailinglist.can_browse_archive(subscribed) or request.user.is_staff ):
        raise Http404("not allowed to browse archives")
    if not month in month_numbers:
        raise Http404("bad month number")
    # security check
    if ".." in path:
        raise Http404("Internal error: list_archives_month :  .. in the path")

    # handle last argument
    archive_index = ARCHIVE_DIR+"/%s/%s/%s/html/" % (str(mailinglist), year, month)
    if path=='':
        path = "threads.html"
        month_number = None
    else:
        month_number = month
    if not path.endswith(".html"):
        # this is something else : just serve it as-is
        # TODO : directly tell webserver to serve it. problem is that it
        # depends on webserver (apache, nginx)..
        from django.views.static import serve
        return serve(request, path, document_root=archive_index)

    # ok, so we have an html file to include
    archive_index += path
    try:
        stat(archive_index)
    except OSError:
        # does not exist..
        raise Http404("Internal error: list_archives_month: can't stat the achive_index")
    # display it
    return render(request, 'archives_month.html', {
        'mailinglist': mailinglist,
        'year' : year,
        'month' : month_name[int(month)],
        'month_number' : month_number,
        'archive_index':archive_index,
    })

@receiver(user_activated)
def login_on_activation(sender, user, request, **kwargs):
    "Listen for activation signal and login the user on success"
    user.backend = 'django.contrib.auth.backends.ModelBackend'
    login(request, user)

#
# Manage user account, all of those require login
#

@login_required
def myaccount(request):
    if request.method == 'POST':
        form = EmailForm(request.POST)
        #check the form
        if form.is_valid():
            return add_secondary_email(request, form.cleaned_data['email'])
    else:
        form = EmailForm()
    profile = request.user.profile
    subs = profile.subscription_set.filter(mailinglist__state=LIST_STATE_PRODUCTION)
    return render(request, 'myaccount.html', {
        'form':form,
        'secondary_emails': profile.secondary_emails.all(),
        'moderator_on': subs.filter(moderator=True),
        'mylists':profile.own_lists.filter(state=LIST_STATE_PRODUCTION).all(),
        'subscriptions': subs,
    })

# not a view, merely a subcase
def add_secondary_email(request, email):
    profile = request.user.profile
    otherprofile = find_email(email, checkConfirmed=False)
    if otherprofile == profile:
        raise asdf
        return render(request, "error.html", {
            "message": "This email address is already used.",
            "returnpage" : reverse(myaccount),
        })
    if otherprofile !=None:
        return render(request, "error.html", { "message":
        """There is already an account using this email. You can
        not add it to your secondary emails list.
        """, "returnpage" : reverse(myaccount)})

    secondary = OtherEmail(profile=profile,email=email)
    secondary.save()
    try:
        secondary.send_confirmation_mail()
    except:
        # ignore SMTP error
        pass
    messages.info(request, "You just added the secondary email %s and a confirmation email was just sent. The email won't be used until confirmation." % secondary.email)
    return HttpResponseRedirect(reverse(myaccount))

@login_required
def remove_secondary_email(request, email_id):
    secondary = get_object_or_404(OtherEmail, id=email_id)
    email = secondary.email
    secondary.delete()
    messages.success(request, "You've just removed the secondary email %s." % email)
    return HttpResponseRedirect(reverse(myaccount))

@login_required
def confirm_secondary_email(request, email_id):
    secondary = get_object_or_404(OtherEmail, id=email_id)
    secondary.send_confirmation_mail()
    messages.info(request, "The confirmation mail has been sent, please check your email.")
    return HttpResponseRedirect(reverse(myaccount))

@login_required
def switch_secondary_email(request, email_id):
    secondary = get_object_or_404(OtherEmail, id=email_id)
    # check
    if not secondary.confirmed:
        raise Http404("Unconfirmed email, how did you get there ?")
    # actual switch
    request.user.email , secondary.email = secondary.email, request.user.email
    request.user.save()
    secondary.save()
    return HttpResponseRedirect(reverse(myaccount))

# secret link without (yet) authentication, no need for @login_required
def activate_secondary_email(request, activation_key):
    """
    Mostly a copy/paste from django-registration activate()
    """
    # first find the key
    activation_key = activation_key.lower() # Normalize before trying anything with it.
    key = get_object_or_404(OtherEmailKey, activation_key=activation_key)

    # if connected... check the user
    # TODO

    # actual activation
    other = key.otheremail
    other.confirmed = True
    other.save()

    # clean
    OtherEmailKey.objects.filter(otheremail=other).delete()

    # go to the myaccount page where otheremails are displayed
    messages.success(request, "The secondary email %s is now confirmed." % other.email)
    return HttpResponseRedirect(reverse(myaccount))

@login_required
def list_manage(request, name):
    mailinglist = getListFromName(name) # or 404
    profile = request.user.profile

    # check we are the owner
    if mailinglist.owner!=profile and not request.user.is_staff:
        raise Http404

    if request.method == 'POST':
        form = ListForm(request.POST, instance=mailinglist)
        #check the form
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(reverse(list_manage, args=[str(mailinglist),]))
    else:
        form = ListForm(instance=mailinglist)
    return render(request, 'list_manage.html', { 'form':form, })

# vim: ai ts=4 sts=4 et sw=4
